#!/usr/bin/env python3
"""
Windows Audio Manager (System Management Tool)

Parameters:
    --list [TYPE]: List audio devices (playback/recording)
    --get-volume DEVICE_ID: Get device volume
    --set-volume DEVICE_ID LEVEL: Set device volume (0-100)
    --mute DEVICE_ID: Mute device
    --unmute DEVICE_ID: Unmute device
    --sessions: List active audio sessions
    --format FORMAT: Output format (table/json)

Requires: pycaw, win32com, tabulate packages
Warning: Some operations may require admin privileges
"""

import sys
import json
import logging
from typing import List, Dict, Optional, Union
from ctypes import cast, POINTER
from comtypes import CLSCTX_ALL

from tabulate import tabulate
from pycaw.pycaw import (
    AudioUtilities, 
    IAudioEndpointVolume, 
    ISimpleAudioVolume,
    AudioDevice
)

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)

class AudioManager:
    """Manages Windows audio devices and settings."""
    
    def __init__(self):
        """Initialize the AudioManager."""
        self.devices = AudioUtilities.GetAllDevices()
        
    def get_audio_devices(self, device_type: Optional[str] = None) -> List[Dict]:
        """
        Get a list of audio devices.
        
        Args:
            device_type: Optional filter for device type ('playback' or 'recording')
            
        Returns:
            List[Dict]: List of audio devices with their properties
        """
        devices = []
        
        for device in self.devices:
            dev_type = "playback" if device.state == 1 else "recording"
            if device_type and dev_type != device_type.lower():
                continue
                
            devices.append({
                "id": device.id,
                "name": device.FriendlyName,
                "type": dev_type,
                "state": "active" if device.state == 1 else "disabled",
                "default": self._is_default_device(device)
            })
            
        return devices
    
    def _is_default_device(self, device: AudioDevice) -> bool:
        """
        Check if a device is the default audio device.
        
        Args:
            device: Audio device to check
            
        Returns:
            bool: True if device is default, False otherwise
        """
        default_speakers = AudioUtilities.GetSpeakers()
        default_mic = AudioUtilities.GetMicrophone()
        
        return (device.id == default_speakers.id or 
                device.id == default_mic.id)
    
    def get_device_volume(self, device_id: str) -> Optional[float]:
        """
        Get the volume level of a specific device.
        
        Args:
            device_id: ID of the audio device
            
        Returns:
            Optional[float]: Volume level (0-100) or None if device not found
        """
        device = self._get_device_by_id(device_id)
        if not device:
            logger.error(f"Device with ID {device_id} not found")
            return None
            
        try:
            interface = device.Activate(
                IAudioEndpointVolume._iid_, CLSCTX_ALL, None)
            volume = cast(interface, POINTER(IAudioEndpointVolume))
            return round(volume.GetMasterVolumeLevelScalar() * 100, 1)
        except Exception as e:
            logger.error(f"Error getting volume for device {device_id}: {e}")
            return None
    
    def set_device_volume(self, device_id: str, volume: float) -> bool:
        """
        Set the volume level of a specific device.
        
        Args:
            device_id: ID of the audio device
            volume: Volume level (0-100)
            
        Returns:
            bool: True if successful, False otherwise
        """
        device = self._get_device_by_id(device_id)
        if not device:
            logger.error(f"Device with ID {device_id} not found")
            return False
            
        try:
            interface = device.Activate(
                IAudioEndpointVolume._iid_, CLSCTX_ALL, None)
            volume_interface = cast(interface, POINTER(IAudioEndpointVolume))
            volume_interface.SetMasterVolumeLevelScalar(volume / 100, None)
            logger.info(f"Set volume to {volume}% for device {device.FriendlyName}")
            return True
        except Exception as e:
            logger.error(f"Error setting volume for device {device_id}: {e}")
            return False
    
    def mute_device(self, device_id: str, mute: bool = True) -> bool:
        """
        Mute or unmute a specific device.
        
        Args:
            device_id: ID of the audio device
            mute: True to mute, False to unmute
            
        Returns:
            bool: True if successful, False otherwise
        """
        device = self._get_device_by_id(device_id)
        if not device:
            logger.error(f"Device with ID {device_id} not found")
            return False
            
        try:
            interface = device.Activate(
                IAudioEndpointVolume._iid_, CLSCTX_ALL, None)
            volume = cast(interface, POINTER(IAudioEndpointVolume))
            volume.SetMute(mute, None)
            action = "muted" if mute else "unmuted"
            logger.info(f"Successfully {action} device {device.FriendlyName}")
            return True
        except Exception as e:
            logger.error(f"Error setting mute state for device {device_id}: {e}")
            return False
    
    def get_audio_sessions(self) -> List[Dict]:
        """
        Get a list of active audio sessions.
        
        Returns:
            List[Dict]: List of audio sessions with their properties
        """
        sessions = []
        
        try:
            for session in AudioUtilities.GetAllSessions():
                if session.Process:
                    volume_interface = session._ctl.QueryInterface(ISimpleAudioVolume)
                    sessions.append({
                        "name": session.Process.name(),
                        "pid": session.Process.pid,
                        "volume": round(volume_interface.GetMasterVolume() * 100, 1),
                        "muted": volume_interface.GetMute()
                    })
        except Exception as e:
            logger.error(f"Error getting audio sessions: {e}")
            
        return sessions
    
    def set_session_volume(self, pid: int, volume: float) -> bool:
        """
        Set the volume level for a specific audio session.
        
        Args:
            pid: Process ID of the audio session
            volume: Volume level (0-100)
            
        Returns:
            bool: True if successful, False otherwise
        """
        try:
            for session in AudioUtilities.GetAllSessions():
                if session.Process and session.Process.pid == pid:
                    volume_interface = session._ctl.QueryInterface(ISimpleAudioVolume)
                    volume_interface.SetMasterVolume(volume / 100, None)
                    logger.info(f"Set volume to {volume}% for process {session.Process.name()}")
                    return True
            logger.error(f"No audio session found for PID {pid}")
            return False
        except Exception as e:
            logger.error(f"Error setting session volume for PID {pid}: {e}")
            return False
    
    def _get_device_by_id(self, device_id: str) -> Optional[AudioDevice]:
        """
        Get an audio device by its ID.
        
        Args:
            device_id: ID of the audio device
            
        Returns:
            Optional[AudioDevice]: Audio device or None if not found
        """
        for device in self.devices:
            if device.id == device_id:
                return device
        return None

def main():
    """Main function to handle command-line interface."""
    import argparse
    
    parser = argparse.ArgumentParser(description="Windows Audio Manager")
    parser.add_argument("--list-devices", action="store_true", 
                      help="List audio devices")
    parser.add_argument("--type", choices=["playback", "recording"],
                      help="Filter device type")
    parser.add_argument("--get-volume", metavar="DEVICE_ID",
                      help="Get volume level for specified device")
    parser.add_argument("--set-volume", nargs=2, metavar=("DEVICE_ID", "VOLUME"),
                      help="Set volume level (0-100) for specified device")
    parser.add_argument("--mute", metavar="DEVICE_ID",
                      help="Mute specified device")
    parser.add_argument("--unmute", metavar="DEVICE_ID",
                      help="Unmute specified device")
    parser.add_argument("--list-sessions", action="store_true",
                      help="List active audio sessions")
    parser.add_argument("--set-session-volume", nargs=2, 
                      metavar=("PID", "VOLUME"),
                      help="Set volume level (0-100) for specified session")
    parser.add_argument("--format", choices=["table", "json"], default="table",
                      help="Output format (table or json)")
    
    args = parser.parse_args()
    
    audio_manager = AudioManager()
    
    try:
        if args.list_devices:
            devices = audio_manager.get_audio_devices(args.type)
            if args.format == "json":
                print(json.dumps(devices, indent=2))
            else:
                headers = ["ID", "Name", "Type", "State", "Default"]
                table_data = [[dev["id"], dev["name"], dev["type"], 
                             dev["state"], dev["default"]] for dev in devices]
                print(tabulate(table_data, headers=headers, tablefmt="grid"))
                
        elif args.get_volume:
            volume = audio_manager.get_device_volume(args.get_volume)
            if volume is not None:
                print(f"Volume: {volume}%")
                
        elif args.set_volume:
            device_id, volume = args.set_volume
            if audio_manager.set_device_volume(device_id, float(volume)):
                print(f"Successfully set volume to {volume}%")
                
        elif args.mute:
            if audio_manager.mute_device(args.mute, True):
                print("Device muted successfully")
                
        elif args.unmute:
            if audio_manager.mute_device(args.unmute, False):
                print("Device unmuted successfully")
                
        elif args.list_sessions:
            sessions = audio_manager.get_audio_sessions()
            if args.format == "json":
                print(json.dumps(sessions, indent=2))
            else:
                headers = ["Process", "PID", "Volume", "Muted"]
                table_data = [[s["name"], s["pid"], f"{s['volume']}%", 
                             s["muted"]] for s in sessions]
                print(tabulate(table_data, headers=headers, tablefmt="grid"))
                
        elif args.set_session_volume:
            pid, volume = args.set_session_volume
            if audio_manager.set_session_volume(int(pid), float(volume)):
                print(f"Successfully set session volume to {volume}%")
                
        else:
            parser.print_help()
            
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        sys.exit(1)

if __name__ == "__main__":
    main()
